from typing import Optional
import platform
import rich.box
import logging
import os
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Confirm, Prompt
from .info import BaseInfo, LinuxInfo, WindowsInfo, DarwinInfo, BSDInfo
from .operating_system import OperatingSystem
from smah.console import std_console, err_console, prompt_choice, prompt_string

def operating_system_terminal_configurator(operating_system: Optional[OperatingSystem]) -> OperatingSystem:
    std_console.print(Markdown("### Configure Operating System"))
    operating_system = operating_system or OperatingSystem()

    if operating_system.is_configured():
        show(operating_system)
        if not Confirm.ask("edit?", default=False):
            return operating_system

    while True:
        operating_system = prompt(operating_system)

        std_console.print("\n")
        show(operating_system, label = "Confirm Operating System Settings", border_style="green bold")
        if Confirm.ask("confirm?", default=True):
            return operating_system

def details():
    os_type = None
    release = None
    version = None
    name = None
    try:
        os_type = platform.system()
        release = platform.release()
        version = platform.version()
        name = os.name
    except Exception as e:
        logging.error("Exception {e}", str(e))

    if os_type == "Darwin":
        os_type = "Darwin (macOs)"

    return {
        "type": os_type,
        "release": release,
        "version": version,
        "name": name,
    }


def prompt(operating_system: OperatingSystem) -> OperatingSystem:
    d = details()
    if Prompt.ask("Automatically Load Operating System Details"):
        operating_system.type = d["type"]
        operating_system.name = d["name"]
        operating_system.version = d["version"]
        operating_system.release = d["release"]
        # TODO - load existing
        operating_system.info = load_info(d["type"])
    else:
        operating_system.type = prompt_choice("Type", operating_system.type or d.get("type"), options=["Linux", "Windows", "Darwin (macOs)", "FreeBSD", "OpenBSD", "NetBSD"], other=True)
        operating_system.name = prompt_string("Name", operating_system.name or d.get("name"))
        operating_system.version = prompt_string("Version", operating_system.version or d.get("version"))
        operating_system.release = prompt_string("Release", operating_system.release or d.get("release"))
        # TODO - load existing
        operating_system.info = load_info(d["type"])
    return operating_system

def load_info(os_type: str) -> Optional[BaseInfo]:
    if os_type == "Linux":
        return LinuxInfo(fetch=True)
    elif os_type == "Windows":
        return WindowsInfo(fetch=True)
    elif os_type == "Darwin (macOs)":
        return DarwinInfo(fetch=True)
    elif "BSD" in os_type:
        return BSDInfo(fetch=True)
    else:
        return BaseInfo(os_type, fetch=True)

def show(operating_system: OperatingSystem, label = "Operating System Settings", border_style="white bold"):
    body = operating_system.show()
    template = Markdown(body)
    template = Panel(template, title=label, border_style=border_style, box=rich.box.ROUNDED)
    std_console.print(template)


# smah/settings/system/info/__init__.py
from .base_info import BaseInfo
from .bsd_info import BSDInfo
from .darwin_info import DarwinInfo
from .linux_info import LinuxInfo
from .windows_info import WindowsInfo

__all__ = ['BaseInfo', 'BSDInfo', 'DarwinInfo', 'LinuxInfo', 'WindowsInfo']


from typing import Optional, Tuple

from .base_info import BaseInfo
import subprocess

class WindowsInfo(BaseInfo):
    """
    Represents Windows-specific OS details.

    Attributes:
        errors (list): List of errors encountered during initialization.
        source (str): The source of the OS details.
        details (dict): The OS details.
        vsn (str): Version string.
        configured (bool): Indicates if the details are configured.
    """

    @staticmethod
    def kind():
        """
        Returns the kind of the OS details.

        Returns:
            str: The kind of the OS details.
        """
        return "Windows"

    @staticmethod
    def info() -> Optional[Tuple]:
        return WindowsInfo.system_info_details()

    @staticmethod
    def system_info_details() -> Optional[Tuple]:
        """
        Retrieves OS details using the systeminfo command.

        Returns:
            tuple: A tuple containing the source and details dictionary.
        """
        try:
            version_info = subprocess.check_output(['systeminfo']).decode('utf-8')
            version_dict = {}
            for line in version_info.split('\n'):
                parts = line.strip().split(':')
                if len(parts) == 2:
                    key, value = parts
                    key = key.strip().replace(' ', '-').lower()
                    value = value.strip()
                    version_dict[key] = value
            return "system info", version_dict
        except (subprocess.CalledProcessError, FileNotFoundError):
            return None

    def __init__(self, config_data = None, fetch=False):
        """
        Initializes the WindowsDetails instance with the given configuration data.

        Args:
            config_data (dict): Configuration data for the OS details.
            fetch (bool): Whether to fetch the details.
        """
        super().__init__("Windows", config_data=config_data, fetch=fetch)



import textwrap

import yaml

from smah.settings import Settings
from smah.settings.inference import Inference


class Prompts:
    MAX_PIPE_LENGTH = 2048
    PIPE_HEAD_LENGTH = 1024

    def __init__(self):
        pass

    @staticmethod
    def run_command_tool():
        return {
            "type": "function",
            "function": {
                "name": "run_command",
                "description": "Runs a specified command on the system and returns the output.",
                "strict": True,
                "parameters": {
                    "type": "object",
                    "properties": {
                        "command": {
                            "type": "string",
                            "description": "The command string to be executed in the shell."
                        },
                        "title": {
                            "type": "string",
                            "description": "A descriptive title for the command execution."
                        },
                        "purpose": {
                            "type": "string",
                            "description": "The reason or objective for running this command."
                        }
                    },
                    "additionalProperties": False,
                    "required": ["command", "title", "purpose"]
                }
            }
        }

    @staticmethod
    def planner_response_format():
        """
        Returns the format for the model picker.

        Returns:
            dict: A dictionary containing the format for the model picker.
        """

        return {
            "type": "json_schema",
            "json_schema": {
                "name": "model-pick",
                "strict": True,
                "schema": {
                    "type": "object",
                    "properties": {
                        "title": {
                          "type": "string",
                          "description": "Concise title describing user's request."
                        },
                        "model": {
                            "type": "string",
                            "description": "The model to use for the response.",
                        },
                        "reason": {
                            "type": "string",
                            "description": "The reason for selecting the model.",
                        },
                        "include_settings": {
                            "type": "boolean",
                            "description": "Whether to include system settings for request. e.g. if system specific settings are needed to tailor response to user's system.",
                        },
                        "include_settings_reason": {
                            "type": "string",
                            "description": "The reason for including system settings.",
                        },
                        "format_output": {
                            "type": "boolean",
                            "description": "Whether to format the output (true) or output raw text (false).",
                        },
                        "format_output_reason": {
                            "type": "string",
                            "description": "The reason for formatting_output choice",
                        },
                        "instructions": {
                            "type": "string",
                            "description": "Additional instructions to pass the model that will be respond to request, set to None if you have no instructions to add"
                        }
                    },
                    "additionalProperties": False,
                    "required": ["title", "model", "reason", "include_settings", "include_settings_reason","format_output","format_output_reason", "instructions"],
                }
            }
        }


    @staticmethod
    def message(role="user", content="..."):
        """
        Generates simple text message.

        Args:
            role (str): The role of the message.
            content (str): The content of the acknowledgment message.

        Returns:
            dict: A dictionary containing the role and content of the acknowledgment message.
        """
        return {
            "role": role,
            "content": content
        }

    @staticmethod
    def ack(role="assistant", content="ack"):
        """
        Generates an acknowledgment message with the given content.

        Args:
            role (str): The role of the acknowledgment message.
            content (str): The content of the acknowledgment message.

        Returns:
            dict: A dictionary containing the role and content of the acknowledgment message.
        """
        return Prompts.message(role=role, content=content)



    @staticmethod
    def conventions():
        template = textwrap.dedent(
            """
            Noizu Prompt Lingua
            ========
            The following NPL prompt conventions will be used in this conversation.

            # Conventions
            - `highlight`: emphasize key terms.
            - `in-fill`: `[...]`, `[...<size>]` indicates sections to be filled in with generated content.
              - Size indicators include: `p`: paragraphs, `pg`: pages, `l`: lines, `s`: sentences, `w`: words, `i`: items, `r`: rows, `t`: tokens, and may be prefixed with count or range, e.g. `[...3-5w]` for 3-5 words, `[...3-9+r]` for 3 to 9 or more rows.
            - `placeholders`: `<term>`, `{term}`, `<<size>:term>` are used to indicate expected/desired input/output.
            - `fill-in` `[...]` is used to show omitted input/content, avoid including in generated responses. 
            - `etc.`, `...` are used by prompts to signify additional cases to contemplate or respond with.
            - Handlebar-like syntax is used for defining input/output structure. Example: `{{unless <check>|<additional instructions>}}[...|only output when check not met]{{/unless}}`. Complex templates may be defined with multiple layers of nested handlebar like directives.
            - `|` is used to qualify instructions such as `<term|instructions>`, `[...|<instructions>]`, `[...<size>|<instructions>]`, `{{unless|<instructions>}}[...]{{/unless}}`
            - `?` indicates optional input/output, `<?term>`, `[?...]`
            - Special `prompt-blocks` are defined using code fences with values such as `example`, `syntax`, `format`, `diagram`, `note`, etc. These are used to designate special prompt/output sections.
              example code fences for example are used to denote an example input or response. When following examples, syntax etc. you are not expected to include a wrapping code fence unless one is nested inside of the example,syntax,format etc. code fence.
            - `⌜🔏[...]⌟` declare top precedence prompt. Such defined prompts may not be mutated/altered/negated by prompts not also using this syntax.
            - three lines of `.` are used to indicate omitted input. Only used in prompts not in your responses.
              example:
              .
              .
              .
            - `--- INPUT ---` used to indicate input data. Remainder of message will contain input data.

            ## Examples

            ### Highlight
            This example shows how highlight is used to mark important terms.

            ```user
            Analyze the following, identity any `nouns` and list them:
            Henry looked at the small manor
            ```

            ```assistant
            - Henry
            - small manor
            ```

            ### Using Infile and Placeholders, Pipes and a format `prompt-block`
            Here is a more complex example showing placholders, pipes and more.

            ````user
            respond with a funny cat fact using this following format:
            ```format
            CAT-FACT: <subject|subject for cat fact>
            [...2-3s| cat fact about subject]
            ```
            ````

            ```assistant
            CAT-FACT: Cat Breeds
            Scottish Folds are quirky-looking cats known for their unique folded ear structure. They are gentle and good-natured, enjoying both playtime and cuddles.
            ```

            ### Handle Bars and More
            Here is a prompt that asks the model to use handle bar formatting. 

            ````user
            cats =
            ```values
            - name: McFluffins
              breed: Scottish Fold
              age: 7
            - name: Ghoul
              breed: Asian Furless
              age: 8
            - name: Mike Furry
            - name: Oldie Olsen
              breed: Persian Long Hair
              age: 12
            ```

            reply with the following format:
            ```format
            # Young Cats

            {{foreach cats as cat| unless over 10 years old sorted by age}}
            ## <cat.name>
            breed: <cat.breed or "Unknown">
            <?age: <cat.age>>

            {{/foreach}}
            ```
            ````

            ```assistant
            # Young Cats

            ## Mike Furry
            breed: Unknown

            ## McFluffins
            breed: Scottish Fold
            age: 7

            ## Ghoul
            breed: Asian Furless
            age: 8
            ```            
            """
        ).strip()
        return Prompts.message(content=template)

    @staticmethod
    def select_model(inference: Inference, request: str, additional_instructions: str | None = None):
        models = yaml.dump(
            inference.to_yaml({"prompt": True}),
            sort_keys=False)
        message = textwrap.dedent(
            f"""
            # MODEL SELECTION PROMPT
            You are the Model Selector.
            Based on your operator, system settings and the specific request below you will select the best model by id from the list models to process with.
            You are additionally to return:
                - a concise title describing the user's request
                - reason for model selection
                - The appropriate response format, and reason for selection
                - if system settings are required, and reason for selection. If user is requesting commands to run for example you will need to include system settings.
                - additional instructions for handling their request.

            Weigh cost and speed in selecting your model, generally cheaper and faster is best unless the problem is highly complicated and requires a slower more advanced model
                                                
            {additional_instructions}

            ## Models
            ```yaml
            {models}
            ```
            ---
            Request
            ===
            {request}            
            """).format(models=models, request=request, additional_instructions=additional_instructions)
        return Prompts.message(content=message)

    @staticmethod
    def pipe_request(request: str, pipe: str):
        if len(pipe) > Prompts.MAX_PIPE_LENGTH:
            pipe_head = pipe[:Prompts.PIPE_HEAD_LENGTH]
            pipe_tail = pipe[Prompts.PIPE_HEAD_LENGTH:]
            r = textwrap.dedent(
                """
                {request}
                --- INPUT ---
                {pipe_head}
                .
                .
                .
                {pipe_tail}
                """
            ).format(request=request, pipe_head=pipe_head, pipe_tail=pipe_tail)
        else:
            r = textwrap.dedent(
                """
                {request}
                --- INPUT ---
                {pipe}                
                """
            ).format(request=request, pipe=pipe)
        return r

    @staticmethod
    def system_settings(settings: Settings, include_system=True):
        """
        Generates a system settings prompt based on the provided settings.
        """
        operator = yaml.dump(settings.user.to_yaml({"stats": True, "prompt": True}), sort_keys=False)
        system = yaml.dump(settings.system.to_yaml({"prompt": True, "stats": True}), sort_keys=False)
        if not include_system:
            template = textwrap.dedent(
                """
                Settings
                ================
                Review and Reply ack.

                # Operator
                About your operator.

                ```yaml
                {operator}
                ```
                """).strip().format(operator=operator)
        else:
            template = textwrap.dedent(
                """
                Settings
                ================
                Review and Reply ack.

                # Operator
                About your operator.

                ```yaml
                {operator}
                ```

                # System
                About this system.
                ```yaml
                {system}
                ```
                """).strip().format(operator=operator, system=system)
        return Prompts.message(content=template)

    @staticmethod
    def query_prompt(request: str):
        prompt = textwrap.dedent(
            """
            # PROMPT
            You are an in-terminal AI Assistant.
            You are to be helpful and take into account your operator's experience levels and details in responding to all requests.
            Open your response with an Intent Statement

            ## Chain of Thought
            Use inline cot statements to improve the output of your response. Every paragraph should open or close with at least one thought statement. 
            
            Thought Statements:
            Note use the backtick symbol to open and close your thought statements.
            - Use `<cot type="thinking">[...|statement]</cot>` before inside of and after output sections to plan your output and improve it.
            - Use `<cot type="assumption">[...|statement]</cot>` before output sections to clarify assumptions made in your reply.
            - Use `<cot type="tangent">[...|statement]</cot>` before, inside and after output sections to reflect on tangential items that may improve your output or are interesting.
            - Use `<cot type="inner-critic">[...|statement]</cot>` after output sections to reflect on your output and improve it. If indicated you can proceed with additional output based on your inner critic's feedback. 

            ## Intent Statements
            Intent statements are statements where you plan out how you will reply they should be placed inside an intent code-fence.            
            ````example
            ```intent
            List the pros and cons of [...| item user requested] taking into account [...|detail about user and system].
            [...| further planning statements]
            ```
            ````
            
            ## User Executable Commands
            in your output if providing commands for user to run wrap those commands in a exec tag. Do not wrap in a shell code fence.
            
            ```example
            <exec shell="zsh">
            <title>Save Hello World to tmp file</title>
            <purpose>Save Hello World to tmp</purpose>
            <command>
            echo "Hello World" > ./tmp
            </command>
            </exec>
            ```
        

            ## Closing Remarks
            Your output should be heavily interspersed with thinking comments and your output should adjust based on observations made in those thought comments.
            Such as identifying an error or issue.                 
            ----
            # User Inquiry
            {request}
            """
        ).format(request=request)
        return Prompts.message(content=prompt)

    @staticmethod
    def pipe_prompt():
        prompt = textwrap.dedent(
            """
            # SYSTEM PROMPT
            You are AI assisted Input processor. Your operator provides an instructions for processing input data and you return output that can be passed to additional terminal programs.

            **Important!** PLEASE Do not output anything other what was requested. Adding comments before or after the processed input may break when passing your output to other command line programs.

            ## Examples

            lets say operator asked you to extract some values from the provided input here is a valid response:                
            ```example
            12
            435
            ```

            Meanwhile below is an invalid response example, it is invalid because of the comments before and after the generated output.
            ```example
            Sure let me extract those values for you.

            12
            435

            I hope that helps!
            ```
            
            DO NOT WRAP YOUR RESPONSE TO THE PIPE OUTPUT IN A CODE FENCE
            ---
            When you are ready, reply ack.
            """)
        return Prompts.message(content=prompt)


from datetime import datetime
from typing import Optional

import yaml

def get_list(source: dict, value: str) -> Optional[list[str]]:
    v = source.get(value)
    if isinstance(v, str):
        v = [v]
    return v

def get_iso(source: dict, value: str) -> Optional[datetime]:
    v = source.get(value)
    return datetime.fromisoformat(v) if isinstance(v, str) else None

class ModelUseCase:
    def __init__(self, config_data = None):
        config_data = config_data or {}
        self.name: str = config_data.get("name")
        self.instructions: Optional[str] = config_data.get("instructions")
        self.notes: Optional[str] = config_data.get("notes")
        self.score: float = config_data.get("score", 0.3)

    def to_yaml(self, options=None):
        o = {
            'name': self.name,
        }
        if self.instructions:
            o['instructions'] = self.instructions
        if self.notes:
            o['notes'] = self.notes
        if self.score:
            o['score'] = self.score
        return o

    def show(self, options = None):
        y = self.to_yaml(options=None)
        return yaml.dump(y, sort_keys=False)

class Model:
    CONFIG_VSN: str = "0.0.1"

    @staticmethod
    def config_vsn() -> str:
        """
        Returns the version of the configuration.

        Returns:
            str: The version of the configuration.
        """
        return Model.CONFIG_VSN

    def __init__(self, provider, config_data):
        self.provider: str = provider
        self.vsn: str = config_data.get("vsn")

        self.name: str = config_data.get("name")
        self.model: str = config_data.get("model")
        self.description: str = config_data.get("description")
        self.enabled: bool = config_data.get("enabled", False)
        self.training_cutoff: Optional[datetime] = get_iso(config_data, "training_cutoff")
        self.license: Optional[str] = config_data.get("license")
        self.model_type: Optional[str] = config_data.get("model_type", "LLM")
        self.context: Optional[dict] = config_data.get("context")
        self.strengths: list[str] = get_list(config_data, "strengths") or []
        self.weaknesses: list[str] = get_list(config_data, "weaknesses") or []
        self.modalities: Optional[dict] = config_data.get("modalities") or {}
        self.settings: Optional[dict] = config_data.get("settings") or {}
        self.attributes: Optional[dict] = config_data.get("attributes") or {}
        self.cost: Optional[dict] = config_data.get("cost")
        v = config_data.get("use_cases") or []
        self.use_cases: Optional[list] = []
        for use_case in v:
            self.use_cases.append(ModelUseCase(use_case))

    def is_configured(self):
        if not self.name:
            return False
        if not self.model:
            return False
        if self.enabled is None:
            return False
        return True

    def to_yaml(self, options=None):
        options = options or {}
        if options.get("prompt"):
            return self.to_prompt_yaml(options=options)
        else:
            return self.to_standard_yaml(options=options)


    def to_prompt_yaml(self, options=None):
        options = options or {}
        min_score = options.get('min_score', 0.4)
        use_cases = [x.to_yaml(options=options) for x in self.use_cases if x.score >= min_score]
        o = {
            'id': f"{self.provider}.{self.name}",
            'description': self.description,
            'training_cutoff': self.training_cutoff,
            'context': self.context,
            'strengths': self.strengths,
            'weaknesses': self.weaknesses,
            'modalities': self.modalities,
            'attributes': self.attributes,
            'cost': self.cost,
            'use_cases': use_cases,
        }
        return o

    def to_standard_yaml(self, options=None):
        options = options or {}
        if options.get("save"):
            use_cases = [x.to_yaml(options=options) for x in self.use_cases]
        else:
            min_score = options.get('min_score', 0.4)
            use_cases = [x.to_yaml(options=options) for x in self.use_cases if x.score >= min_score]

        o = {
            'name': self.name,
            'model': self.model,
            'description': self.description,
            'enabled': self.enabled,
            'training_cutoff': self.training_cutoff,
            'license': self.license,
            'model_type': self.model_type,
            'context': self.context,
            'strengths': self.strengths,
            'weaknesses': self.weaknesses,
            'modalities': self.modalities,
            'settings': self.settings,
            'attributes': self.attributes,
            'cost': self.cost,
            'use_cases': use_cases,
            'vsn': self.config_vsn(),
        }
        if options.get("save"):
            pass
        elif options.get("disabled"):
            o.pop('vsn')
            o.pop('license')
            o.pop('settings')
        else:
            o.pop('vsn')
            o.pop('enabled')
            o.pop('license')
            o.pop('settings')
        return o

    def show(self, options = None):
        y = self.to_yaml(options=options)
        return yaml.dump(y, sort_keys=False)


import textwrap
from typing import Optional

from .info import LinuxInfo, DarwinInfo, WindowsInfo, BSDInfo, BaseInfo


class OperatingSystem:
    CONFIG_VSN: str = "0.0.1"

    @staticmethod
    def config_vsn() -> str:
        """
        Returns the version of the configuration.

        Returns:
            str: The version of the configuration.
        """
        return OperatingSystem.CONFIG_VSN

    @staticmethod
    def version_supported(version) -> bool:
        """
        Checks if the given version is supported.

        Args:
            version (str): The version to check.

        Returns:
            bool: True if the version is supported, False otherwise.
        """
        if version is None:
            return False
        else:
            return version <= OperatingSystem.config_vsn()

    @staticmethod
    def load_info(config_data: Optional[dict]) -> Optional[BaseInfo]:
        """
        Loads OS-specific details from the configuration data.

        Args:
            config_data (dict): The configuration data.

        Returns:
            object: The OS-specific details.
        """
        config_data = config_data or {}
        kind = config_data.get("kind")
        if kind == "Linux":
            return LinuxInfo(config_data)
        elif kind == "Darwin (macOs)":
            return DarwinInfo(config_data)
        elif kind == "Windows":
            return WindowsInfo(config_data)
        elif kind == "BSD":
            return BSDInfo(config_data)
        elif kind:
            return BaseInfo(kind, config_data)
        else:
            return None

    def __init__(self, config_data = None):
        config_data = config_data or {}
        self.type: Optional[str] = config_data.get("type")
        self.name: Optional[str] = config_data.get("name")
        self.version: Optional[str] = config_data.get("version")
        self.release: Optional[str] = config_data.get("release")
        self.info: Optional[BaseInfo] = self.load_info(config_data.get("info"))
        self.vsn: Optional[str] = config_data.get("vsn")

    def is_configured(self):
        """
        Checks if the system is configured.

        Returns:
            bool: True if the system is configured, False otherwise.
        """
        if not self.type:
            return False
        if not self.name:
            return False
        if not self.version:
            return False
        if not self.release:
            return False
        if not self.info or not self.info.is_configured():
            return False
        if not self.vsn:
            return False
        return True

    def to_yaml(self, options = None):
        """
        Converts the system configuration to a YAML object.

        Args:
            options (dict): The options to include in the YAML object.

        Returns:
            dict: The system configuration as a YAML object.
        """
        return {
            "type": self.type,
            "name": self.name,
            "version": self.version,
            "release": self.release,
            "info": self.info.to_yaml(options=options) if self.info else None,
            "vsn": self.config_vsn()
        }

    def show(self, options = None):
        info = textwrap.indent(self.info.show(options=options), "  ") if self.info else None
        template = textwrap.dedent(
            """
            - type: {type}
            - name: {name}
            - version: {version}
            - release: {release}
            - info:
            {info}
            """
        ).strip().format(
            type=self.type,
            name=self.name,
            version=self.version,
            release=self.release,
            info=info
        )
        return template



from typing import Optional, Tuple

from .base_info import BaseInfo
import subprocess

class LinuxInfo(BaseInfo):
    """
    Represents Linux-specific OS details.

    Attributes:
        errors (list): List of errors encountered during initialization.
        source (str): The source of the OS details.
        details (dict): The OS details.
        vsn (str): Version string.
        configured (bool): Indicates if the details are configured.
    """

    @staticmethod
    def kind():
        """
        Returns the kind of the OS details.

        Returns:
            str: The kind of the OS details.
        """
        return "Linux"

    @staticmethod
    def info() -> Optional[Tuple]:
        return LinuxInfo.os_release_details() or LinuxInfo.uname_details() or LinuxInfo.proc_details()

    @staticmethod
    def os_release_details() -> Optional[Tuple]:
        """
        Retrieves OS release details from /etc/os-release.

        Returns:
            tuple: A tuple containing the source and details dictionary.
        """

        try:
            # Attempt to parse /etc/os-release first
            distro_info = {}
            with open('/etc/os-release', 'r') as f:
                lines = f.readlines()
            for line in lines:
                parts = line.strip().split('=')
                if len(parts) == 2:
                    key = parts[0].strip().lower().replace('_', '-')
                    value = parts[1].strip().strip('"')
                    distro_info[key] = value
            return "os-release", distro_info
        except (FileNotFoundError, PermissionError):
            return None

    @staticmethod
    def uname_details() -> Optional[Tuple]:
        """
        Retrieves OS details using the uname command.

        Returns:
            tuple: A tuple containing the source and details dictionary.
        """

        try:
            uname_info = subprocess.check_output(['uname', '-a']).decode('utf-8').strip().split()
            distro_info = {'os_type': uname_info[0], 'os_release': uname_info[2], 'os_version': uname_info[3]}
            return "uname", distro_info
        except (subprocess.CalledProcessError, FileNotFoundError, IndexError):
            return None

    @staticmethod
    def proc_details() -> Optional[Tuple]:
        """
        Retrieves OS details from /proc/version.

        Returns:
            tuple: A tuple containing the source and details dictionary.
        """
        try:
            # Fallback to parsing /proc/version
            with open('/proc/version', 'r') as f:
                version_info = f.readline().strip().split()
            distro_info = {'os_type': version_info[0], 'os_release': version_info[2], 'os_version': version_info[4]}
            return "proc", distro_info
        except (FileNotFoundError, PermissionError, IndexError):
            return None

    def __init__(self, config_data=None, fetch=False):
        """
        Initializes the LinuxDetails instance with the given configuration data.

        Args:
            config_data (dict): Configuration data for the OS details.
            fetch (bool): Whether to fetch the details.
        """
        super().__init__("Linux", config_data=config_data, fetch=fetch)


from xml.etree.ElementTree import Element

from rich.console import Console, RenderResult, JustifyMethod, ConsoleOptions
import textwrap

from rich.markdown import Markdown
from rich.markdown import MarkdownElement
from rich.prompt import Prompt, Confirm
from typing import Optional, Union

from rich.segment import Segment
from rich.style import Style

# Initialize standard console for general output
std_console: Console = Console(legacy_windows=False)

# Initialize error console for error output
err_console: Console = Console(legacy_windows=False, stderr=True)

def prompt_string(
        field: str,
        value: Optional[str],
        required: bool = True,
        label: Optional[str] = None,
        style: Optional[str] = "green bold"
        ) -> str:
    if value:
        std_console.print(f"{field.capitalize()}: {value}")
        if not Confirm.ask("edit?", default=False):
            return value
    msg = f"[{style}]{label or field.capitalize()}[/{style}]"

    v = Prompt.ask(msg, default=value)
    while required and not v:
        v = Prompt.ask(msg, default=value)
    return v

def prompt_choice(
    field: str,
    value: Optional[str],
    options: list[str] | list[tuple[str, str]],
    required: bool = True,
    other: bool = False,
    label: Optional[str] = None,
    style: Optional[str] = "green bold") -> str:
    if value:
        std_console.print(f"{field.capitalize()}: {value}")
        if not Confirm.ask("edit?", default=False):
            return value

    lookup = {}
    reverse_lookup = {}
    i = 0
    for option in options:
        if isinstance(option, tuple):
            k, v = option
        else:
            i += 1
            k = str(i)
            v = option
        lookup[k] = v
        reverse_lookup[v] = k
    if other:
        lookup["o"] = "Other"

    o = "\n".join([f"{k} - {v}" for k, v in lookup.items()])
    o = textwrap.indent(o, " ")
    msg = f"[{style}]{label or field.capitalize()}[/{style}]\n{o}\n"

    d = reverse_lookup.get(value)
    d  = d or "o" if other and value else None

    choice = Prompt.ask(msg, choices=list(lookup.keys()), default=d)
    if other and choice == "o":
        v = Prompt.ask(f"[{style}]Other[/{style}]", default=value)
        while required and not v:
            v = Prompt.ask(f"[{style}]Other[/{style}]", default=value)
        return v

    return lookup[choice]


import argparse
import sys
import logging

def merge_args(args: argparse.Namespace, config: dict) -> argparse.Namespace:
    """
    Merges parsed command-line arguments with configuration settings.

    Args:
        args (Namespace): Parsed command-line arguments.
        config (dict): Configuration settings.

    Returns:
        Namespace: Merged arguments and settings.
    """
    for key, value in config.items():
        if key in args and getattr(args, key) is None:
            setattr(args, key, value)
    return args

def extract_args() -> tuple[argparse.Namespace, str | None]:
    """
    Parses and extracts command-line arguments for the SMAH CLI tool.

    Returns:
        parser (ArgumentParser): The argument parser with configured options.
        args (Namespace): Parsed arguments and options.
        pipe (str or None): Content read from standard input if available.
    """
    parser = __initialize_argument_parser()
    __add_general_arguments(parser)
    __add_ai_arguments(parser)
    __add_gui_arguments(parser)
    a = parser.parse_args()
    return a, __get_pipe()

def __initialize_argument_parser():
    """
    Initialize the argument parser with basic configurations.

    Returns:
        ArgumentParser: A configured argument parser.
    """
    return argparse.ArgumentParser(
        description="SMAH Command Line Tool",
        formatter_class=argparse.RawTextHelpFormatter
    )


def __add_general_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Add general command-line arguments to the parser.

    Args:
        parser (ArgumentParser): The argument parser to which general arguments are added.
    """
    parser.add_argument('-q', '--query', type=str, help='The Query to process')
    parser.add_argument('-i', '--instructions', type=str, help='The Instruction File to process')
    parser.add_argument('--interactive', action=argparse.BooleanOptionalAction, help='Run in interactive mode', default=False)
    parser.add_argument('-c', '--config', type=str, help='Path to alternative config file')
    parser.add_argument('--database', type=str, help='Path to sqlite smah database')
    parser.add_argument('--configure', action=argparse.BooleanOptionalAction, help='Enter Config Setup', default=False)
    parser.add_argument('--continue', dest="resume", action=argparse.BooleanOptionalAction, help='Continue Last Conversation', default=False)
    parser.add_argument('--session', type=int, help='Resume Session')
    parser.add_argument('--history', action=argparse.BooleanOptionalAction, help='Resume Recent Session', default=False)
    parser.add_argument('-v', '--verbose', action='count', default=0, help="Set Verbosity Level, such as -vv")

def __add_ai_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Add specific AI-related command-line arguments to the parser.

    Args:
        parser (ArgumentParser): The argument parser to which AI-related arguments are added.
    """
    parser.add_argument('--model', type=str, help='Default Model')
    parser.add_argument('--model-picker', type=str, help='Picker Model')
    parser.add_argument('--model-query', type=str, help='Query Model')
    parser.add_argument('--model-pipe', type=str, help='Pipe Processing Model')
    parser.add_argument('--model-interactive', type=str, help='Interactive Processing Model')
    parser.add_argument('--model-review', type=str, help='Output Reviewer Model')
    parser.add_argument('--model-edit', type=str, help='Output Editor Model')

    parser.add_argument('--openai-api-tier', type=int, help='OpenAI Tier')
    parser.add_argument('--openai-api-key', type=str, help='OpenAI Api Key')
    parser.add_argument('--openai-api-org', type=str, help='OpenAI Api Org')

def __add_gui_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Add GUI-related command-line arguments to the parser.

    Args:
        parser (ArgumentParser): The argument parser to which GUI-related arguments are added.
    """
    parser.add_argument('--gui', action=argparse.BooleanOptionalAction, help='Run in GUI mode', default=False)
    parser.add_argument('--rich', action=argparse.BooleanOptionalAction, help='Rich Format Output', default=True)

def __get_pipe():
    """
    Reads data from standard input if present and available.

    Returns:
        str or None: The content available from standard input; otherwise, None if input is not a TTY.
    """
    if sys.stdin.isatty():
        return None
    else:
        return sys.stdin.read()

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    args, pipe = extract_args()
    logging.debug("Parsed arguments: %s", args)
    logging.debug("Pipe content: %s", pipe)